import numpy as np
from dataclasses import dataclass


@dataclass
class Tank:
    """
    Class that holds the parameters of the tank, providing a method for
    checking if a position is inside the tank.

    Args:
        - radius (float): radius of the tank.
        - height (float): height of the tank.
        - position (np.ndarray): position of the tank, default [0, 0, 0].
    """

    radius: float
    height: float
    position: np.ndarray

    def __post_init__(self):
        """
        Calculate the volume and energy loss fraction.
        """
        self.volume = np.pi * self.radius**2 * self.height

    def inside(self, position: np.ndarray) -> bool:
        """
        Check if a position is inside the tank.

        Args:
            - position (np.ndarray): position to check.

        Returns: True if the position is inside the tank, False otherwise.
        """
        return bool(
            np.linalg.norm(position[:2] - self.position[:2]) < self.radius
            and abs(position[2] - self.position[2]) < self.height / 2
        )



import numpy as np

from neutrowater.diffusing_neutrons import DiffusingNeutrons, Vector


class Measurer:
    """Class to measure the properties of the neutrons after the simulation."""

    def __init__(self, sim: DiffusingNeutrons):
        self.neutrons = sim.neutrons
        self.sim = sim

    def positions(self) -> list[list[Vector]]:
        """
        Get the positions of the neutrons.

        Returns a list where each element is a list of length nNeutrons:
        [[np.ndarray, np.ndarray, ...], [np.ndarray, np.ndarray, ...], ...]
        """
        return [neutron.positions for neutron in self.neutrons]

    def energies(self) -> list[list[float]]:
        """
        Get the energies of the neutrons.

        Returns a list where each element is a list of length nNeutrons:
        [[float, float, ...], [float, float, ...], ...]
        """
        return [neutron.energies for neutron in self.neutrons]

    def number_total(self) -> int:
        """
        Get the total number of neutrons.

        Returns an int.
        """
        return len(self.neutrons)

    def number_escaped(self) -> int:
        """
        Get the number of neutrons that escaped the tank.

        Returns an int.
        """
        return sum(
            [
                1
                for neutron in self.neutrons
                if not self.sim.tank.inside(neutron.position)
            ],
            start=0,
        )

    def number_thermal(self):
        """
        Get the number of neutrons that are thermalized.

        Returns an int.
        """
        n = 0
        for energies in self.energies():
            if energies[-1] < (10 * self.sim.kT):
                n += 1
        return n

    def thermalize_positions(self) -> list:
        """
        Get the positions of the neutrons where their energy is thermal for
        the first time.
        """
        thermalize_positions = []
        for energies, positions in zip(self.energies(), self.positions()):
            # Get the first element of the list of energies that is less than 10kT
            res = list(filter((lambda val: val < (10 * self.sim.kT)), energies))
            if len(res) > 0:
                index = energies.index(res[0])
                thermalize_positions.append(positions[index])

        return thermalize_positions

    def thermalize_distances(self) -> list:
        """
        Get the distance of the neutrons where their energy is thermal for
        the first time.
        """
        return [np.linalg.norm(pos) for pos in self.thermalize_positions()]

    def number_absorbed(self) -> int:
        """
        Get the number of neutrons that were absorbed.

        Returns an int.
        """
        n = 0
        for neutron in self.neutrons:
            if len(neutron.positions) < self.sim.nCollisions and self.sim.tank.inside(
                neutron.position
            ):
                n += 1
        return n

    def absorbed_positions(self) -> list[Vector]:
        """
        Get the positions of the neutrons that were absorbed.

        Returns a list of np.ndarray.
        """
        return [
            neutron.positions[-1]
            for neutron in self.neutrons
            if len(neutron.positions) < self.sim.nCollisions
            and self.sim.tank.inside(neutron.position)
        ]

    def absorbed_distances(self) -> list[np.float64]:
        """
        Get the distances of the neutrons that were absorbed.

        Returns a list of floats.
        """
        return [np.linalg.norm(pos) for pos in self.absorbed_positions()]

    def flux(self, r: float) -> float:
        """
        Get the flux of the neutrons at a given radius r.

        Args:
            - r (float): radius at which to compute the flux.

        Returns a float value of the flux.
        """
        count = 0
        for positions in self.positions():
            positions = np.array(positions)
            norms = np.linalg.norm(positions, axis=1)
            # Count the number of neutrons that have been in the shell
            count += np.sum((norms[:-1] < r) & (r < norms[1:])) + np.sum(
                (norms[1:] < r) & (r < norms[:-1])
            )
        return count / (4 * np.pi * r**2)

    def energy_spectrum(self, r: float) -> list[float]:
        """
        Get the energy spectrum of the neutrons at a given radius r.

        Args:
            - r (float): radius at which to compute the energy spectrum.

        Returns a list of floats.
        """
        result_energies = []
        for positions, energies in zip(self.positions(), self.energies()):
            positions = np.array(positions)
            energies = np.array(energies)
            norms = np.linalg.norm(positions, axis=1)
            # Create boolean arrays for the conditions
            cond1 = np.concatenate([(norms[:-1] < r) & (r < norms[1:]), [False]])
            cond2 = np.concatenate([[False], (norms[1:] < r) & (r < norms[:-1])])
            # Get the energies of the neutrons that have been in the shell
            result_energies += list(energies[cond1]) + list(energies[cond2])
        return result_energies

    def energy_spectrum_escaped(self) -> list[float]:
        """
        Get the energy spectrum of the neutrons that escaped the tank.

        Returns a list of floats.
        """
        return [
            neutron.energy
            for neutron in self.neutrons
            if not self.sim.tank.inside(neutron.position)
        ]

    def number_above_energy(self, E: float) -> int:
        """
        Get the number of neutrons with energy above a given energy.

        Args:
            - E (float): energy threshold.

        Returns an int.
        """
        return sum([1 for neutron in self.neutrons if neutron.energies[0] > E])



import numpy as np
from typing import Sequence, overload
from collections.abc import Iterator
import typing


Vector = np.ndarray[typing.Any, np.dtype[np.float64]]


class Neutron:
    """
    Class that holds the parameters of the Neutron, providing methods for
    traveling and colliding, which adjust the relevant class parameters.
    """

    def __init__(self, initial_energy: float, initial_position: Vector):
        self.positions: list = [initial_position]
        self.energies: list = [initial_energy]

    @property
    def position(self) -> Vector:
        return self.positions[-1]

    @property
    def energy(self) -> float:
        return self.energies[-1]

    @position.setter
    def position(self, vector: Vector) -> None:
        self.positions.append(vector)

    @energy.setter
    def energy(self, value: float) -> None:
        self.energies.append(value)

    def travel(self, distance: float):
        """
        Update the position by letting the neutron travel in a given direction
        for a given distance.
        """
        self.position = self.position + distance * self.direction

    def collide(self, energy_loss_frac: float, direction: Vector):
        """
        Update the energy by letting the neutron collide resulting in a given
        energy loss.
        """
        self.direction = direction
        self.energy = self.energy * energy_loss_frac


class Neutrons:
    """Class that holds multiple neutrons and their parameters."""

    def __init__(
        self,
        energies: Sequence[float],
        positions: Sequence[Vector],
    ):
        """
        energies (Sequence[float]): initial energies of the neutrons, should be
            same length as positions.
        positions (Sequence[np.ndarray[np.float64]]): initial positions of the
            neutrons, should be same length as energies.
        """
        self.energies = energies
        self.positions = positions
        self._init_neutrons()

    def _init_neutrons(self):
        """
        Initialize the neutrons with an initial energy and position
        """
        self.neutrons = [
            Neutron(energy, position)
            for energy, position in zip(self.energies, self.positions)
        ]

    def __iter__(self) -> Iterator[Neutron]:
        """
        Allows for iterating over the neutrons in the class.
        """
        return iter(self.neutrons)

    def __len__(self) -> int:
        """
        Returns the number of neutrons in the class.
        """
        return len(self.neutrons)

    @overload
    def __getitem__(self, index: int) -> Neutron: ...

    @overload
    def __getitem__(self, index: slice) -> list[Neutron]: ...

    def __getitem__(self, index: int | slice) -> Neutron | list[Neutron]:
        """
        Returns the neutron at a given index.
        """
        return self.neutrons[index]



import numpy as np
from numpy.random import random
import pandas as pd
from typing import Sequence, Any
import multiprocessing
from tqdm.contrib.concurrent import process_map
from dataclasses import dataclass

from neutrowater.models.neutrons import Neutrons, Neutron, Vector
from neutrowater.models.tank import Tank
from neutrowater.models.collisions import Collision, random_direction
from neutrowater.process.data_processor import (
    TotalProcessor,
    SpectrumProcessor,
    AbsorptionProcessor,
)
from neutrowater.process.angular_processor import AngularProcessor
from neutrowater.models.maxwell_boltzmann import MaxwellBoltzmann


@dataclass
class Parameters:
    """
    Parameters:
    - nNeutrons (int): Number of neutrons to simulate
    - molecule_structure (Sequence): number of atoms in the molecule (Default:
        (2, 1) for H20)
    - nuclei_masses (Sequence): Masses of the nuclei in the molecule, should be
        in the same order as the molecule_structure. (Default: (1, 16) for H20)
    - radius_tank (float): Radius of the tank in meters. Default is 1.
    - height_tank (float): Height of the tank in meters. Default is 1.
    - position_tank (np.ndarray): Position of the tank. Default is [0, 0, 0].
    - temperature (float): Temperature [K] of the medium. Default is 293.
    """

    nNeutrons: int
    molecule_structure: Sequence = (2, 1)
    nuclei_masses: Sequence = (1, 16)
    radius_tank: float = 1
    height_tank: float = 1
    position_tank: Vector = np.array([0.0, 0.0, 0.0])
    temperature: float = 293

    def __post_init__(self):
        self.total_data: Sequence[pd.DataFrame] = [
            pd.read_csv("../data/h_cross_t.txt", sep=r"\s+"),
            pd.read_csv("../data/o_cross_t.txt", sep=r"\s+"),
        ]
        self.scattering_data: Sequence[pd.DataFrame] = [
            pd.read_csv("../data/h_cross_s.txt", sep=r"\s+"),
            pd.read_csv("../data/o_cross_s.txt", sep=r"\s+"),
        ]
        self.absorption_data: Sequence[pd.DataFrame] = [
            pd.read_csv("../data/h_cross_a.txt", sep=r"\s+"),
            pd.read_csv("../data/o_cross_a.txt", sep=r"\s+"),
        ]
        self.angular_data: Sequence[pd.DataFrame] = [
            pd.read_csv("../data/H_angular.txt", sep=r"\s+"),
            pd.read_csv("../data/O_angular.txt", sep=r","),
        ]
        self.spectrum_data: pd.DataFrame = pd.read_csv(
            "../data/neutron_spectrum_normalized.txt", sep=","
        )


class DiffusingNeutrons:
    """
    Class that simulates multiple neutrons from diffusing in a medium.

    Args:
        - p (Parameters): Parameters for the simulation.
    """

    def __init__(
        self,
        p: Parameters,
    ):
        self.kT = (
            1.380649e-23 * p.temperature * 6.24150907 * 10**18
        )  # [m^2 kg / (s^2 K) * K * (eV/J) = J*(eV/J) = eV]
        self.nCollisions = 0
        self.mol_struc = p.molecule_structure
        self.tank = Tank(p.radius_tank, p.height_tank, p.position_tank)

        # For interpolating the total cross section data and computing the
        # mean-free-path.
        self.total_processor = TotalProcessor(p.total_data)

        # For interpolating the scattering and absorption cross section and
        # computing the aborption ratio.
        self.absorption_processor = AbsorptionProcessor(
            p.scattering_data, p.absorption_data
        )

        # Sample from the interpolated energy spectrum
        spectrum_processor = SpectrumProcessor(p.spectrum_data)
        initial_energies = spectrum_processor.sample(num_samples=p.nNeutrons)

        # Convert the energies MeV -> eV
        initial_energies = [energy * 10**6 for energy in initial_energies]

        # All neutrons start at the origin
        initial_positions = [np.array([0, 0, 0]) for _ in range(p.nNeutrons)]
        self.neutrons = Neutrons(initial_energies, initial_positions)

        # For handling the collisions with the atomic nuclei, hydrogen and oxygen.
        self.nuclei_masses = p.nuclei_masses

        # Maxwell-Boltzmann distribution for the thermal energy
        self.mw = MaxwellBoltzmann(T=p.temperature)

        # For handling the angular distribution of the scattering.
        self.angular_processor = AngularProcessor(p.angular_data, p.nuclei_masses)

    def diffuse(self, nCollisions: int):
        """
        Let the neutrons diffuse in the medium.

        Args:
            - nCollisions (int): number of times each neutron collides with an
            atomic nucleus.
        """
        self.nCollisions += nCollisions
        # Using not all cores but 2 less than the total number of cores, tends
        # to be faster
        num_processes = multiprocessing.cpu_count() - 2

        # Split the neutrons into chunks
        chunk_size = (len(self.neutrons) + num_processes - 1) // num_processes
        chunks = [
            self.neutrons[i : i + chunk_size]
            for i in range(0, len(self.neutrons), chunk_size)
        ]

        # Diffuse the chuncks of neutrons in parallel
        with multiprocessing.Pool(processes=num_processes) as pool:
            results = list(
                process_map(self._diffuse_chunk, chunks, [nCollisions] * len(chunks)),
            )

        # Update the neutrons list with the results from the parallel processess
        self.neutrons = [neutron for chunk in results for neutron in chunk]

    def _diffuse_chunk(self, chunk: Neutrons, nCollisions: int):
        """
        Diffuse a chunk of all neutrons in the medium.

        Args:
            - chunk (Sequence[Neutron]): chunk of neutrons to diffuse
            - nCollisions (int): number of times each neutron collides with an atomic
            nucleus.

        Returns: the neutrons after diffusing in the medium.
        """
        np.random.seed()  # Set a new seed for each process
        return [self._diffuse_neutron(neutron, nCollisions) for neutron in chunk]

    def _diffuse_neutron(self, neutron: Neutron, nCollisions: int) -> Neutron:
        """
        Diffuse a single neutron in the medium.

        Args:
            - neutron (Neutron): Neutron to diffuse
            - nCollisions (int): number of times the neutron collides with an atomic
            nucleus.

        Returns: the neutron after diffusing in the medium.
        """
        neutron.direction = random_direction()

        for _ in range(nCollisions):

            if not self.tank.inside(neutron.position):
                break

            # Sample the distance the neutron travels from an exponential distribution
            # with mean free path as the scale parameter.
            l = -self.total_processor.get_mfp(neutron.energy) * np.log(random())
            neutron.travel(l)

            # Determine the nucleus the neutron collides with
            mass = (
                self.nuclei_masses[0]
                if random() < self.total_processor.get_ratio(neutron.energy)
                else self.nuclei_masses[1]
            )

            collision = Collision(
                initial_E=neutron.energy,
                initial_direction=neutron.direction,
                mass=mass,
                scattering_cosine=self.angular_processor.get_CM_cosines(
                    mass, neutron.energy, 1
                )[0],
                absorption=self._absorbed(neutron, self.nuclei_masses.index(mass)),
                thermal=neutron.energy < 10 * self.kT,
            )

            # Update the neutron's energy and direction
            neutron.collide(collision.energy_loss_frac, collision.scattering_direction)

            # If the neutron is absorbed, break the loop
            if collision.energy_loss_frac == 1:
                break

        return neutron

    def _absorbed(self, neutron: Neutron, index: int) -> bool:
        """
        Check if the neutron is absorbed by the nucleus.

        Args:
            - neutron (Neutron): Neutron to check.
            - index (int): index of the nucleus in the molecule.

        Returns: True if the neutron is absorbed, False otherwise.
        """
        if (
            random()
            < self.absorption_processor.get_absorption_rates(neutron.energy)[index]
        ):
            return True
        else:
            return False



import pandas as pd
import numpy as np
from typing import Sequence


class AngularProcessor:
    """
    Class that holds the angular distribution data for neutrons in water and
    provides a method for sampling the angular distribution for a given energy.
    """

    def __init__(self, data: Sequence[pd.DataFrame], masses: Sequence[float]):
        """
        Args:
            - data (Sequence[pd.DataFrame]): angular distribution data of the form
            of a sequence of pandas DataFrames with the first column energy and the
            rest of the columns the coefficients of the Legendre series.
            - masses (Sequence[float]): list of atomic masses of the nuclei in the medium.
        """

        self.data = data
        self.bounds = (-1, 1)
        self.masses = masses

    def _legendre_series(self, data_coeffs):
        """
        Evaluate the coefficients of the Legendre series. The coefficients are
        multiplied by the factor (2 * (i + 1) + 1) / 2, where i is the index of
        the coefficient.

        Args:
            - coeffs (list): coefficients of the Legendre series in the data.

        Returns: list of coefficients of the Legendre series.
        """
        coeffs = [coef * (2 * (i + 1) + 1) / 2 for i, coef in enumerate(data_coeffs)]
        # Add the 0th coefficient.
        coeffs.insert(0, 0.5)
        return coeffs

    def _distribution_index(self, E: float, j: int):
        """
        Find the index of the angular distribution for a given energy.

        Args:
            - E (float): energy of the neutron in eV.
            - j (int): index of the mass in the list

        Returns: index of the distribution.
        """

        # Find the index of the energy in the data closest to the given energy
        energies = self.data[j].iloc[:, 0].to_list()
        for i, val in enumerate(energies):
            if val > E:
                break

        # Use interpolation factor to decide which index to use
        f = (E - energies[i - 1]) / (energies[i] - energies[i - 1])

        if np.random.random() > f:
            return i - 1
        else:
            return i

    def get_CM_cosines(
        self,
        mass: float,
        E: float,
        num_samples: int,
    ) -> list:
        """
        Sample the spectrum using Monte Carlo sampling.

        Args:
            - mass (float): mass of the nucleus.
            - E (float): energy of the neutron in eV.
            - num_samples (int): number of samples to take.

        Returns: list of sampled cosines of the scattering angle.
        """
        j = self.masses.index(mass)

        # Find the index of the distribution.
        index = self._distribution_index(E, j)

        # Get the coefficients of the Legendre series with the index.
        coeffs = self._legendre_series(self.data[j].iloc[index, :].tolist()[1:])

        # Sample the cosines of the scattering angle.
        sampled_points = []
        while len(sampled_points) < num_samples:
            x = np.random.uniform(*self.bounds)
            y = np.random.uniform(0, 1)

            px = np.polynomial.legendre.legval(x, coeffs)

            if y <= px:
                sampled_points.append(x)

        return sampled_points



import numpy as np


# J -> eV
J = 6.24150907e18  # eV

# Boltzmann constant
kB = 1.3806452e-23


class MaxwellBoltzmann:
    """
    Class to sample from the Maxwell-Boltzmann distribution.
    """

    def __init__(self, m: float = 1.67493e-27, T: float = 293):
        """
        Args:
            - m (float): mass of the particle in kg.
            - T (float): temperature in K.
        """
        self.m = m
        self.T = T

    def distribution(self, v: float) -> float:
        """
        Maxwell-Boltzmann distribution for velocities.

        Args:
            - v (float): velocity
        """
        return (
            (self.m / (2 * np.pi * kB * self.T)) ** 1.5
            * 4
            * np.pi
            * v**2
            * np.exp(-self.m * v**2 / (2 * kB * self.T))
        )

    def thermal_velocity(self) -> float:
        """
        Samples a thermal velocity of the particle.

        Returns: thermal velocity in m/s.
        """
        return self._sample()[0]

    def thermal_energy(self) -> float:
        """
        Sample a thermal energy of the particle.

        Returns: thermal energy in eV.
        """
        return 0.5 * self.m * self.thermal_velocity() ** 2 * J

    def _sample(
        self,
        x_min: float = 0,
        x_max: float = 10000,
        y_max: float = 0.0004,
        num_samples: int = 1,
    ) -> list:
        """
        Sample from a function using Monte Carlo sampling.

        Args:
            - f (Callable): function to sample from.
            - x_min (float): minimum x value.
            - x_max (float): maximum x value.
            - y_max (float): maximum y value.
            - num_samples (int): number of samples to take.

        Returns: list of sampled points.
        """
        sampled_points = []
        while len(sampled_points) < num_samples:
            x = np.random.uniform(x_min, x_max)
            y = np.random.uniform(0, y_max)

            px = self.distribution(x)

            if y <= px:
                sampled_points.append(x)

        return sampled_points



import pandas as pd
import numpy as np
from scipy.interpolate import Akima1DInterpolator
from typing import Sequence
import itertools


class DataProcessor:
    """
    Class that holds data and provides a method for interpolating. The data is
    of the form of a sequence of pandas DataFrames with columns x and y.

    Attributes:
    - interpolaters (Sequence[Akima1DInterpolator]): interpolaters for the data.
    """

    def __init__(self, data: Sequence[pd.DataFrame], log: bool) -> None:
        """
        interpolate the data using scipy.interpolate.Akima1DInterpolator.

        Args:
            - data (Sequence[pd.DataFrame]): cross section data of the form of a
            sequence of pandas DataFrames with columns x and y.
            - log (bool): if True, the data is log-log transformed.
        """
        self.interpolaters = [self._interpolate(d, log) for d in data]

    @staticmethod
    def _interpolate(data: pd.DataFrame, log: bool = False) -> Akima1DInterpolator:
        """
        Preprocess the data to be used for the mean free path calculations.

        Args:
            - data (pd.DataFrame): cross section data of the form of a pandas
            DataFrame with columns x and y.
            - log (bool): if True, the data is log-log transformed.
        """
        data = data.drop_duplicates(subset=data.columns.values.tolist()[0])
        data = data.loc[(data != 0.0).all(axis=1)]

        x = data.iloc[:, 0]
        y = data.iloc[:, 1]
        if log:
            x = np.log(x)
            y = np.log(y)

        return Akima1DInterpolator(x, y)


class CrossSectionProcessor(DataProcessor):
    """
    Class that holds the cross section data for neutrons in water and provides a
    method for calculating the mean free path for a given energy. The data is of
    the form of a sequence of pandas DataFrames with columns x and y.

    Attributes:
    - interpolaters (Sequence[Akima1DInterpolator]): interpolaters for the data.
    """

    def __init__(self, data: Sequence[pd.DataFrame]):
        """
        Args:
            - data (Sequence[pd.DataFrame]): cross section data of the form of a
            sequence of pandas DataFrames with columns x and y.
        """
        super().__init__(data=data, log=True)

    def cross_section(
        self,
        energy: float,
        f: Akima1DInterpolator | None = None,
    ) -> float:
        """
        Get the cross section for a given energy in m^2.

        Args:
            - energy (float): energy of the neutron in eV.
            - f (Akima1DInterpolator): interpolater for the cross section data.
        """
        f = self.interpolaters[0] if f is None else f
        return np.exp(f(np.log(energy))) * 10 ** (-28)  # convert barns -> m^2


class TotalProcessor(CrossSectionProcessor):
    """
    Class that holds the total cross section data for neutrons and provides a
    method for calculating the mean free path for a given energy.
    """

    def get_mfp(
        self,
        energy: float,
        n: float = 3.3327677048150236e28,
        nMolecules: Sequence = (2, 1),
    ) -> float:
        """
        Get the mean free path in meters for a given energy.

        Args:
            - energy (float): energy of the neutron in eV.
            - n (float): number of atoms per volume in m^-3 (Default is for H20).
            - nMolecules (Sequence): number of atoms in the molecule (Default:
            (2, 1) for H20)
        """
        cross_sections = [self.cross_section(energy, f) for f in self.interpolaters]
        return 1 / (
            sum([nMolecules[i] * n * cs for i, cs in enumerate(cross_sections)])
        )

    def get_ratio(
        self,
        energy: float,
        nMolecules: Sequence = (2, 1),
    ):
        """
        Get the ratio of the total cross section of the first atom to the total
        cross section of the molecule.

        Args:
            - energy (float): energy of the neutron in eV.
            - n (float): number of atoms per volume in m^-3 (Default is for H20).
            - nMolecules (Sequence): number of atoms in the molecule (Default:
            (2, 1) for H20)
        """
        cross_sections = [self.cross_section(energy, f) for f in self.interpolaters]
        return (
            cross_sections[0]
            * nMolecules[0]
            / sum(
                [nMolecules[i] * cross_sections[i] for i in range(len(cross_sections))]
            )
        )


class AbsorptionProcessor(CrossSectionProcessor):
    """
    Class that holds the absorption cross section data for neutrons and provides
    a method for calculating the absorption rate for a given energy.
    """

    def __init__(
        self,
        scattering_data: Sequence[pd.DataFrame],
        absorption_data: Sequence[pd.DataFrame],
    ):
        """
        Args:
            - scattering_data (Sequence[pd.DataFrame]): scattering cross section data
            of the form of a sequence of pandas DataFrames with columns x and y
            - absorption_data (Sequence[pd.DataFrame]): absorption cross section data
            of the form of a sequence of pandas DataFrames with columns x and y.

        Scattering and absorption data should be in the same order of atoms and
        of the same length.
        """
        # List with the scattering and absorption data interleaved.
        data = list(itertools.chain(*zip(scattering_data, absorption_data)))
        super().__init__(data)

    def get_total_absorption_rate(
        self,
        energy,
        nMolecules: Sequence = (2, 1),
    ) -> float:
        """
        Get the absorption rate for a given energy.

        Args:
            - energy (float): energy of the neutron in eV.
            - n (float): number of atoms per volume in m^-3 (Default is for H20).
            - nMolecules (Sequence): number of atoms in the molecule (Default:
            (2, 1) for H20)

        Returns: absorption rate
        """
        # list with the scattering and absorption data interleaved.
        cross_sections = [self.cross_section(energy, f) for f in self.interpolaters]
        return 1 / sum(
            [
                nMolecules[i // 2] * (cross_sections[i] / cross_sections[i + 1])
                for i in range(0, len(cross_sections), 2)
            ]
        )

    def get_absorption_rates(self, energy: float) -> list:
        """
        Get the absorption rates for a given energy.

        Args:
            - energy (float): energy of the neutron in eV.

        Returns: list of absorption rates for each atom.
        """
        cross_sections = [self.cross_section(energy, f) for f in self.interpolaters]
        return [
            cross_sections[i + 1] / cross_sections[i]
            for i in range(0, len(cross_sections), 2)
        ]


class SpectrumProcessor(DataProcessor):
    """
    Class that holds the spectrum data for neutrons and provides a method for
    sampling the spectrum.

    Attributes:
        - interpolaters (Sequence[Akima1DInterpolator]): interpolaters for the data.
    """

    def __init__(self, data: pd.DataFrame):
        """
        Args:
            - data (pd.DataFrame): spectrum data of the form of a pandas DataFrame with
            columns x and y.
            - log (bool): if True, the data is log-log transformed.
        """
        x = data.iloc[:, 0]
        self.bounds = (min(x), max(x))
        super().__init__(data=[data], log=False)

    def sample(
        self,
        num_samples: int,
    ) -> list:
        """
        Sample the spectrum using Monte Carlo sampling.

        Args:
            - num_samples (int): number of samples to take.

        Returns: list of sampled energies.
        """
        f = self.interpolaters[0]
        sampled_points = []
        while len(sampled_points) < num_samples:
            x = np.random.uniform(*self.bounds)
            y = np.random.uniform(0, 1)

            px = f(x)

            if y <= px:
                sampled_points.append(x)

        return sampled_points



import numpy as np
from dataclasses import dataclass

from neutrowater.models.neutrons import Vector
from neutrowater.models.maxwell_boltzmann import MaxwellBoltzmann


@dataclass
class Collision:
    """
    Class to simulate the collisions of neutrons with atomic nuclei.

    Args:
        - initial_E (float): initial energy of the neutron.
        - initial_direction (Vector): initial direction of the neutron in the labframe.
        - mass (float): mass of the nucleus.
        - scattering_cosine (float): cosine of the scattering angle in the CM frame.
        - absorption (bool): flag to indicate if the collision is an absorption.
        - thermal (bool): flag to indicate if the collision is thermal.

    """

    initial_E: float
    initial_direction: Vector
    mass: float
    scattering_cosine: float
    absorption: bool
    thermal: bool

    def __post_init__(self):
        """
        Define the initial conditions of the collision.
        """

        self.mw = MaxwellBoltzmann()

        # initial velocity of neutron in the labframe
        self.v_n = self.initial_direction * np.sqrt(self.initial_E)

        # Velocity of the CM frame
        v_cm = self.v_n / (1 + self.mass)

        # Initial velocity of neutron in the CM frame
        V_n = self.v_n - v_cm

        # Resulting velocity of the neutron in the labframe
        self.v_n_new = v_cm + self._direction_CM() * np.linalg.norm(V_n)

    def _direction_CM(self) -> Vector:
        """
        Get the post-collision scattering direction of the neutron in the CM frame.

        Returns: a np.ndarray of the post-collision scattering direction.
        """

        # Scattering angle in the CM frame
        theta = np.arccos(self.scattering_cosine)

        # Random azimuthal angle
        phi = np.random.uniform(0, 2 * np.pi)

        # Compute direction in the CM frame
        u = np.sin(theta) * np.cos(phi)
        v = np.sin(theta) * np.sin(phi)
        w = np.cos(theta)

        direction = np.array([u, v, w])

        # Normalize the direction for numerical stability
        return direction / np.linalg.norm(direction)

    @property
    def scattering_direction(self) -> Vector:
        """
        Get the post-collision scattering direction of the neutron.

        Returns: a np.ndarray of the post-collision scattering direction.
        """
        if self.thermal:
            return random_direction()
        else:
            return self.v_n_new / np.linalg.norm(self.v_n_new)

    @property
    def energy_loss_frac(self) -> float:
        """
        Get the fraction of energy lost in a collision for the current collision.

        Returns: fraction of energy lost.
        """
        if self.absorption:
            return 1.0
        elif self.thermal:
            return self.mw.thermal_energy() / self.initial_E
        else:
            return float(
                np.linalg.norm(self.v_n_new) ** 2 / np.linalg.norm(self.v_n) ** 2
            )


def random_direction() -> Vector:
    """
    Sample a random 3D direction.

    Returns: a np.ndarray of N 3D (np.ndarray) vectors.
    """
    vec = np.random.normal(size=3)
    return vec / np.linalg.norm(vec, axis=-1)
